#-------------------------------------------------------------------------------
#
#  Define classes for (uni/multi)-variate kernel density estimation.
#
#  Currently, only Gaussian kernels are implemented.
#
#  Copyright 2004-2005 by Enthought, Inc.
#
#  The code has been adapted by Michael Nowotny to work with GPUs
#  using Cocos from the SciPy code available at
#  https://github.com/scipy/scipy/blob/master/scipy/stats/kde.py
#
#  The open source license of the original code is reproduced below:
#
# Copyright (c) 2001-2002 Enthought, Inc.  2003-2019, SciPy Developers.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#-------------------------------------------------------------------------------

# Standard library imports.
import math
from abc import ABC, abstractmethod
from dataclasses import dataclass
from cached_property import cached_property
import numbers
import typing as tp
import warnings

import cocos.numerics as cn
from cocos.numerics.data_types import NumericArray
import cocos.device as cd
from cocos.numerics.numerical_package_selector import select_num_pack

# SciPy imports.
from scipy import linalg, special
from scipy.special import logsumexp
from scipy._lib._util import check_random_state

from numpy import (asarray, atleast_2d, reshape, zeros, newaxis, dot, exp, pi,
                   sqrt, ravel, power, atleast_1d, squeeze, sum, transpose,
                   ones, cov)
import numpy as np

# Local imports.
# # from . import mvn
# from scipy.stats import mvn
# from ._stats import gaussian_kernel_estimate
#
# from scipy.stats._stats import gaussian_kernel_estimate

__all__ = ['gaussian_kde']


def _check_array_at_right_location_and_convert(array,
                                               gpu: bool,
                                               dtype: np.generic = np.float32):
    if isinstance(array, np.ndarray) and gpu:
        array = cn.array(array)

    if isinstance(array, cn.ndarray) and not gpu:
        array = np.array(array)

    if array.dtype != dtype:
        array = array.astype(dtype)

    return array


def ensure_consistent_numeric_arrays(arrays: tp.Iterable[tp.Optional[NumericArray]],
                                     gpu: bool):
    return tuple(_check_array_at_right_location_and_convert(array, gpu)
                 if array is not None
                 else None
                 for array
                 in arrays)


def _verify_and_get_shape_of_datapoints_datavalues_and_evaluation_points(points: NumericArray,
                                                                         values: NumericArray,
                                                                         xi: NumericArray) \
        -> tp.Tuple[int, int, int]:
    n = points.shape[0]

    if points.ndim > 1:
        d = points.shape[1]
    else:
        d = 1
    m = xi.shape[0]

    if values.ndim > 1:
        p = values.shape[1]
    else:
        p = 1

    if p != 1:
        raise ValueError('p != 1 is not supported')

    if xi.shape[1] != d:
        raise ValueError(f"points and xi must have same trailing dim but the shape of xi is {xi.shape}")

    return n, m, d


def gaussian_kernel_estimate_vectorized_whitened(whitening: NumericArray,
                                                 whitened_points: NumericArray,
                                                 values: NumericArray,
                                                 xi: NumericArray,
                                                 norm: float,
                                                 dtype: np.generic,
                                                 gpu: bool) -> NumericArray:
    # print(f'whitened_points.shape = {whitened_points.shape}')
    # print(f'values.shape = {xi.shape}')
    # print(f'xi.shape = {xi.shape}')

    n, m, d = \
        _verify_and_get_shape_of_datapoints_datavalues_and_evaluation_points(points=whitened_points,
                                                                             values=values,
                                                                             xi=xi)
    whitened_points, values, xi, whitening = \
        ensure_consistent_numeric_arrays((whitened_points, values, xi, whitening), gpu)

    num_pack = select_num_pack(gpu)

    whitened_points = whitened_points.astype(dtype, copy=False)
    whitened_xi = num_pack.dot(xi, whitening).astype(dtype, copy=False)
    values = values.astype(dtype, copy=False)

    # Create the result array and evaluate the weighted sum
    whitened_points = whitened_points.reshape((n, 1, d))
    whitened_xi = whitened_xi.reshape((1, m, d))
    residual = whitened_points - whitened_xi
    arg = residual * residual
    del residual
    if d > 1:
        assert arg.shape == (n, m, d)
        arg = num_pack.sum(arg, axis=2)
    else:
        arg = arg.reshape((n, m))
    # print(arg.shape)
    if not gpu:
        assert arg.shape == (n, m)
    arg = num_pack.exp(- 0.5 * arg) * norm
    if not gpu:
        assert arg.shape == (n, m)

    # estimate = num_pack.dot(arg.T, values)
    estimate = (values * arg).sum(axis=0)
    if estimate.ndim > 1:
        estimate = estimate.squeeze()

    if gpu:
        cd.sync()

    return estimate


def gaussian_kernel_estimate_vectorized(points: NumericArray,
                                        values: NumericArray,
                                        xi: NumericArray,
                                        precision: NumericArray,
                                        dtype: np.generic,
                                        gpu: bool = False) \
        -> NumericArray:
    """
    def gaussian_kernel_estimate(points, real[:, :] values, xi, precision)
    Evaluate a multivariate Gaussian kernel estimate.
    Parameters
    ----------
    points : array_like with shape (n, d)
        Data points to estimate from in d dimenions.
    values : real[:, :] with shape (n, p)
        Multivariate values associated with the data points.
    xi : array_like with shape (m, d)
        Coordinates to evaluate the estimate at in d dimensions.
    precision : array_like with shape (d, d)
        Precision matrix for the Gaussian kernel.
    dtype : the result dtype
    gpu : whether to compute the gaussian kernel estimate on the gpu

    Returns
    -------
    estimate : double[:, :] with shape (m, p)
        Multivariate Gaussian kernel estimate evaluated at the input coordinates.
    """
    num_pack = select_num_pack(gpu)
    n, m, d = \
        _verify_and_get_shape_of_datapoints_datavalues_and_evaluation_points(points=points,
                                                                             values=values,
                                                                             xi=xi)

    # n = points.shape[0]
    #
    # if points.ndim > 1:
    #     d = points.shape[1]
    # else:
    #     d = 1
    # m = xi.shape[0]
    #
    # if values.ndim > 1:
    #     p = values.shape[1]
    # else:
    #     p = 1
    #
    # if p != 1:
    #     raise ValueError('p != 1 is not supported')
    #
    # if xi.shape[1] != d:
    #     raise ValueError("points and xi must have same trailing dim")
    # if precision.shape[0] != d or precision.shape[1] != d:
    #     raise ValueError("precision matrix must match data dims")

    points, values, xi, precision = \
        ensure_consistent_numeric_arrays((points, values, xi, precision), gpu)

    # Rescale the data
    whitening = num_pack.linalg.cholesky(precision).astype(dtype, copy=False)
    points = num_pack.dot(points, whitening).astype(dtype, copy=False)
    # xi = num_pack.dot(xi, whitening).astype(dtype, copy=False)
    values = values.astype(dtype, copy=False)

    # Evaluate the normalisation
    norm = (2 * np.pi) ** (- d / 2) * num_pack.prod(num_pack.diag(whitening))

    # # Create the result array and evaluate the weighted sum
    # points = points.reshape((n, 1, d))
    # xi = xi.reshape((1, m, d))
    # residual = points - xi
    # arg = residual * residual
    # del residual
    # if d > 1:
    #     assert arg.shape == (n, m, d)
    #     arg = num_pack.sum(arg, axis=2)
    # else:
    #     arg = arg.reshape((n, m))
    # assert arg.shape == (n, m)
    # arg = num_pack.exp(- 0.5 * arg) * norm
    # assert arg.shape == (n, m)
    #
    # estimate = num_pack.dot(arg.T, values)
    #
    # if gpu:
    #     cd.sync()
    #
    # return estimate.squeeze()

    return gaussian_kernel_estimate_vectorized_whitened(whitening=whitening,
                                                        whitened_points=points,
                                                        xi=xi,
                                                        values=values,
                                                        norm=norm,
                                                        dtype=dtype,
                                                        gpu=gpu)


def gaussian_kernel_estimate(points, values, xi, precision, dtype):
    """
    def gaussian_kernel_estimate(points, real[:, :] values, xi, precision)
    Evaluate a multivariate Gaussian kernel estimate.
    Parameters
    ----------
    points : array_like with shape (n, d)
        Data points to estimate from in d dimenions.
    values : real[:, :] with shape (n, p)
        Multivariate values associated with the data points.
    xi : array_like with shape (m, d)
        Coordinates to evaluate the estimate at in d dimensions.
    precision : array_like with shape (d, d)
        Precision matrix for the Gaussian kernel.
    Returns
    -------
    estimate : double[:, :] with shape (m, p)
        Multivariate Gaussian kernel estimate evaluated at the input coordinates.
    """
    n = points.shape[0]
    d = points.shape[1]
    m = xi.shape[0]
    p = values.shape[1]

    if p != 1:
        raise ValueError('p != 1 is not supported')

    if xi.shape[1] != d:
        raise ValueError("points and xi must have same trailing dim")
    if precision.shape[0] != d or precision.shape[1] != d:
        raise ValueError("precision matrix must match data dims")

    # Rescale the data
    whitening = np.linalg.cholesky(precision).astype(dtype, copy=False)
    points_ = np.dot(points, whitening).astype(dtype, copy=False)
    xi_ = np.dot(xi, whitening).astype(dtype, copy=False)
    values_ = values.astype(dtype, copy=False)

    # Evaluate the normalisation
    norm = (2 * np.pi) ** (- d / 2)
    for i in range(d):
        norm *= whitening[i, i]

    # Create the result array and evaluate the weighted sum
    estimate = np.zeros((m, p), dtype)
    for i in range(n):
        for j in range(m):
            arg = 0
            for k in range(d):
                residual = (points_[i, k] - xi_[j, k])
                arg += residual * residual

            arg = np.exp(-arg / 2) * norm
            for k in range(p):
                estimate[j, k] += values_[i, k] * arg

    return np.asarray(estimate)


@dataclass(frozen=True)
class GaussianKDEInformation:
    points: np.ndarray  # (d, n) shaped array of datapoints
    weights: np.ndarray  # (d, n) shaped array of weights, optional
    dimension: int  # data dimension
    n: int  # number of data points
    neff: float  # effective sample size


CovarianceFactorFunctionType = tp.Callable[[GaussianKDEInformation], float]


SCOTTS_FACTOR_STRING = 'scotts'
SILVERMAN_FACTOR_STRING = 'silverman'


def compute_scotts_factor(kde_info: GaussianKDEInformation) -> float:
    return power(kde_info.neff, -1.0 / (kde_info.dimension + 4))


def compute_silverman_factor(kde_info: GaussianKDEInformation) -> float:
    d = kde_info.dimension
    neff = kde_info.neff
    return power(neff * (d + 2.0) / 4.0, -1.0 / (d + 4))


# class CovarianceFactor(ABC):
#     @abstractmethod
#     def compute_covariance_factor(self, kde_info: GaussianKDEInformation) -> float:
#         pass
#
#
# class ScottsFactor(CovarianceFactor):
#     def compute_covariance_factor(self, kde_info: GaussianKDEInformation) -> float:
#         return power(kde_info.neff, -1.0 / (kde_info.dimension + 4))
#
#
# class SilvermanFactor(CovarianceFactor):
#     def compute_covariance_factor(self, kde_info: GaussianKDEInformation) -> float:
#         d = kde_info.dimension
#         neff = kde_info.neff
#         return power(neff * (d + 2.0) / 4.0, -1.0 / (d + 4))
#
#
# class LambdaCovarianceFactor(CovarianceFactor):
#     def __init__(self, covariance_factor_fun: tp.Callable[[GaussianKDEInformation], float]):
#         self._covariance_factor_fun = covariance_factor_fun
#
#     def compute_covariance_factor(self, kde_info: GaussianKDEInformation) -> float:
#         return self._covariance_factor_fun(kde_info)


class gaussian_kde:
    """Representation of a kernel-density estimate using Gaussian kernels.

    Kernel density estimation is a way to estimate the probability density
    function (PDF) of a random variable in a non-parametric way.
    `gaussian_kde` works for both uni-variate and multi-variate data. It
    includes automatic bandwidth determination. The estimation works best for
    a unimodal distribution; bimodal or multi-modal distributions tend to be
    oversmoothed.

    Parameters
    ----------
    dataset : array_like
        Datapoints to estimate from. In case of univariate data this is a 1-D
        array, otherwise a 2-D array with shape (# of dims, # of data).
    bw_method : str, scalar or callable, optional
        The method used to calculate the estimator bandwidth.  This can be
        'scott', 'silverman', a scalar constant or a callable.  If a scalar,
        this will be used directly as `kde.factor`.  If a callable, it should
        take a `gaussian_kde` instance as only parameter and return a scalar.
        If None (default), 'scott' is used.  See Notes for more details.
    weights : array_like, optional
        weights of datapoints. This must be the same shape as dataset.
        If None (default), the samples are assumed to be equally weighted
    gpu: whether to evaluate the kernel density estimate on the gpu

    Attributes
    ----------
    dataset : ndarray
        The dataset with which `gaussian_kde` was initialized.
    d : int
        Number of dimensions.
    n : int
        Number of datapoints.
    neff : int
        Effective number of datapoints.

        .. versionadded:: 1.2.0
    factor : float
        The bandwidth factor, obtained from `kde.covariance_factor`, with which
        the covariance matrix is multiplied.
    covariance : ndarray
        The covariance matrix of `dataset`, scaled by the calculated bandwidth
        (`kde.factor`).
    inv_cov : ndarray
        The inverse of `covariance`.

    Methods
    -------
    evaluate
    __call__
    integrate_gaussian
    integrate_box_1d
    integrate_box
    integrate_kde
    pdf
    logpdf
    resample

    Notes
    -----
    Bandwidth selection strongly influences the estimate obtained from the KDE
    (much more so than the actual shape of the kernel).  Bandwidth selection
    can be done by a "rule of thumb", by cross-validation, by "plug-in
    methods" or by other means; see [3]_, [4]_ for reviews.  `gaussian_kde`
    uses a rule of thumb, the default is Scott's Rule.

    Scott's Rule [1]_, implemented as `scotts_factor`, is::

        n**(-1./(d+4)),

    with ``n`` the number of data points and ``d`` the number of dimensions.
    In the case of unequally weighted points, `scotts_factor` becomes::

        neff**(-1./(d+4)),

    with ``neff`` the effective number of datapoints.
    Silverman's Rule [2]_, implemented as `silverman_factor`, is::

        (n * (d + 2) / 4.)**(-1. / (d + 4)).

    or in the case of unequally weighted points::

        (neff * (d + 2) / 4.)**(-1. / (d + 4)).

    Good general descriptions of kernel density estimation can be found in [1]_
    and [2]_, the mathematics for this multi-dimensional implementation can be
    found in [1]_.

    With a set of weighted samples, the effective number of datapoints ``neff``
    is defined by::

        neff = sum(weights)^2 / sum(weights^2)

    as detailed in [5]_.

    References
    ----------
    .. [1] D.W. Scott, "Multivariate Density Estimation: Theory, Practice, and
           Visualization", John Wiley & Sons, New York, Chicester, 1992.
    .. [2] B.W. Silverman, "Density Estimation for Statistics and Data
           Analysis", Vol. 26, Monographs on Statistics and Applied Probability,
           Chapman and Hall, London, 1986.
    .. [3] B.A. Turlach, "Bandwidth Selection in Kernel Density Estimation: A
           Review", CORE and Institut de Statistique, Vol. 19, pp. 1-33, 1993.
    .. [4] D.M. Bashtannyk and R.J. Hyndman, "Bandwidth selection for kernel
           conditional density estimation", Computational Statistics & Data
           Analysis, Vol. 36, pp. 279-298, 2001.
    .. [5] Gray P. G., 1969, Journal of the Royal Statistical Society.
           Series A (General), 132, 272

    Examples
    --------
    Generate some random two-dimensional data:

    >>> from scipy import stats
    >>> def measure(n):
    ...     "Measurement model, return two coupled measurements."
    ...     m1 = np.random.normal(size=n)
    ...     m2 = np.random.normal(scale=0.5, size=n)
    ...     return m1+m2, m1-m2

    >>> m1, m2 = measure(2000)
    >>> xmin = m1.min()
    >>> xmax = m1.max()
    >>> ymin = m2.min()
    >>> ymax = m2.max()

    Perform a kernel density estimate on the data:

    >>> X, Y = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
    >>> positions = np.vstack([X.ravel(), Y.ravel()])
    >>> values = np.vstack([m1, m2])
    >>> kernel = stats.gaussian_kde(values)
    >>> Z = np.reshape(kernel(positions).T, X.shape)

    Plot the results:

    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> ax.imshow(np.rot90(Z), cmap=plt.cm.gist_earth_r,
    ...           extent=[xmin, xmax, ymin, ymax])
    >>> ax.plot(m1, m2, 'k.', markersize=2)
    >>> ax.set_xlim([xmin, xmax])
    >>> ax.set_ylim([ymin, ymax])
    >>> plt.show()

    """

    def __init__(self,
                 dataset: NumericArray,
                 bw_method: tp.Optional[tp.Union[CovarianceFactorFunctionType,
                                                 str,
                                                 tp.Callable,
                                                 numbers.Number]] = None,
                 weights: tp.Optional[NumericArray] = None,
                 gpu: bool = False):

        self._num_pack = select_num_pack(gpu)
        self._gpu = gpu
        self.dataset = atleast_2d(asarray(dataset))
        if not self.dataset.size > 1:
            raise ValueError("`dataset` input should have multiple elements.")

        self.d, self.n = self.dataset.shape

        if weights is not None:
            self._weights = atleast_1d(weights).astype(float)
            self._weights /= np.sum(self._weights)
            if self.weights.ndim != 1:
                raise ValueError("`weights` input should be one-dimensional.")
            if len(self._weights) != self.n:
                raise ValueError("`weights` input should be of length n")
            self._neff = 1.0/np.sum(self._weights**2)
        else:
            self._weights = ones(self.n) / self.n

        self._covariance_factor = self._get_covariance_factor_function_from_bandwidth_type(bw_method)
        self._compute_covariance()

    def evaluate(self, points):
        """
        Evaluate the estimated pdf on a set of points.

        Parameters
        ----------
        points : (# of dimensions, # of points)-array
            Alternatively, a (# of dimensions,) vector can be passed in and
            treated as a single point.

        Returns
        -------
        values : (# of points,)-array
            The values at each point.

        Raises
        ------
        ValueError : if the dimensionality of the input points is different than
                     the dimensionality of the KDE.

        """
        points = atleast_2d(asarray(points))

        d, m = points.shape
        if d != self.d:
            if d == 1 and m == self.d:
                # points was passed in as a row vector
                points = reshape(points, (self.d, 1))
                m = 1
            else:
                raise ValueError(f"points have dimension {d}, dataset has dimension {self.d}")

        output_dtype = np.common_type(self.covariance, points)

        if True:
            # result = gaussian_kernel_estimate_vectorized(points=self.dataset.T,
            #                                              values=self.weights[:, None],
            #                                              xi=points.T,
            #                                              precision=self.inv_cov,
            #                                              dtype=output_dtype,
            #                                              gpu=self._gpu)

            result = gaussian_kernel_estimate_vectorized_whitened(
                        whitening=self.whitening,
                        whitened_points=self.whitened_points,
                        values=self.weights[:, None],
                        xi=points.T,
                        norm=self.normalization_constant,
                        dtype=output_dtype,
                        gpu=self._gpu)

            return result
        else:
            result = gaussian_kernel_estimate(points=self.dataset.T,
                                              values=self.weights[:, None],
                                              xi=points.T,
                                              precision=self.inv_cov,
                                              dtype=output_dtype)
            return result[:, 0]

    __call__ = evaluate

    def integrate_gaussian(self, mean, cov):
        """
        Multiply estimated density by a multivariate Gaussian and integrate
        over the whole space.

        Parameters
        ----------
        mean : aray_like
            A 1-D array, specifying the mean of the Gaussian.
        cov : array_like
            A 2-D array, specifying the covariance matrix of the Gaussian.

        Returns
        -------
        result : scalar
            The value of the integral.

        Raises
        ------
        ValueError
            If the mean or covariance of the input Gaussian differs from
            the KDE's dimensionality.

        """
        mean = atleast_1d(squeeze(mean))
        cov = atleast_2d(cov)

        if mean.shape != (self.d,):
            raise ValueError("mean does not have dimension %s" % self.d)
        if cov.shape != (self.d, self.d):
            raise ValueError("covariance does not have dimension %s" % self.d)

        # make mean a column vector
        mean = mean[:, newaxis]

        sum_cov = self.covariance + cov

        # This will raise LinAlgError if the new cov matrix is not s.p.d
        # cho_factor returns (ndarray, bool) where bool is a flag for whether
        # or not ndarray is upper or lower triangular
        sum_cov_chol = linalg.cho_factor(sum_cov)

        diff = self.dataset - mean
        tdiff = linalg.cho_solve(sum_cov_chol, diff)

        sqrt_det = np.prod(np.diagonal(sum_cov_chol[0]))
        norm_const = power(2 * pi, sum_cov.shape[0] / 2.0) * sqrt_det

        energies = sum(diff * tdiff, axis=0) / 2.0
        result = sum(exp(-energies)*self.weights, axis=0) / norm_const

        return result

    def integrate_box_1d(self, low, high):
        """
        Computes the integral of a 1D pdf between two bounds.

        Parameters
        ----------
        low : scalar
            Lower bound of integration.
        high : scalar
            Upper bound of integration.

        Returns
        -------
        value : scalar
            The result of the integral.

        Raises
        ------
        ValueError
            If the KDE is over more than one dimension.

        """
        if self.d != 1:
            raise ValueError("integrate_box_1d() only handles 1D pdfs")

        stdev = ravel(sqrt(self.covariance))[0]

        normalized_low = ravel((low - self.dataset) / stdev)
        normalized_high = ravel((high - self.dataset) / stdev)

        value = np.sum(self.weights*(
                        special.ndtr(normalized_high) -
                        special.ndtr(normalized_low)))
        return value

    # def integrate_box(self, low_bounds, high_bounds, maxpts=None):
    #     """Computes the integral of a pdf over a rectangular interval.
    #
    #     Parameters
    #     ----------
    #     low_bounds : array_like
    #         A 1-D array containing the lower bounds of integration.
    #     high_bounds : array_like
    #         A 1-D array containing the upper bounds of integration.
    #     maxpts : int, optional
    #         The maximum number of points to use for integration.
    #
    #     Returns
    #     -------
    #     value : scalar
    #         The result of the integral.
    #
    #     """
    #     if maxpts is not None:
    #         extra_kwds = {'maxpts': maxpts}
    #     else:
    #         extra_kwds = {}
    #
    #     value, inform = mvn.mvnun_weighted(low_bounds, high_bounds,
    #                                        self.dataset, self.weights,
    #                                        self.covariance, **extra_kwds)
    #     if inform:
    #         msg = ('An integral in mvn.mvnun requires more points than %s' %
    #                (self.d * 1000))
    #         warnings.warn(msg)
    #
    #     return value

    def integrate_kde(self, other):
        """
        Computes the integral of the product of this  kernel density estimate
        with another.

        Parameters
        ----------
        other : gaussian_kde instance
            The other kde.

        Returns
        -------
        value : scalar
            The result of the integral.

        Raises
        ------
        ValueError
            If the KDEs have different dimensionality.

        """
        if other.d != self.d:
            raise ValueError("KDEs are not the same dimensionality")

        # we want to iterate over the smallest number of points
        if other.n < self.n:
            small = other
            large = self
        else:
            small = self
            large = other

        sum_cov = small.covariance + large.covariance
        sum_cov_chol = linalg.cho_factor(sum_cov)
        result = 0.0
        for i in range(small.n):
            mean = small.dataset[:, i, newaxis]
            diff = large.dataset - mean
            tdiff = linalg.cho_solve(sum_cov_chol, diff)

            energies = sum(diff * tdiff, axis=0) / 2.0
            result += sum(exp(-energies)*large.weights, axis=0)*small.weights[i]

        sqrt_det = np.prod(np.diagonal(sum_cov_chol[0]))
        norm_const = power(2 * pi, sum_cov.shape[0] / 2.0) * sqrt_det

        result /= norm_const

        return result

    def resample(self, size=None, seed=None):
        """
        Randomly sample a dataset from the estimated pdf.

        Parameters
        ----------
        size : int, optional
            The number of samples to draw.  If not provided, then the size is
            the same as the effective number of samples in the underlying
            dataset.
        seed : {None, int, `~np.random.RandomState`, `~np.random.Generator`}, optional
            This parameter defines the object to use for drawing random
            variates.
            If `seed` is `None` the `~np.random.RandomState` singleton is used.
            If `seed` is an int, a new ``RandomState`` instance is used, seeded
            with seed.
            If `seed` is already a ``RandomState`` or ``Generator`` instance,
            then that object is used.
            Default is None.
            Specify `seed` for reproducible drawing of random variates.

        Returns
        -------
        resample : (self.d, `size`) ndarray
            The sampled dataset.

        """
        if size is None:
            size = int(self.neff)

        random_state = check_random_state(seed)
        norm = transpose(random_state.multivariate_normal(
            zeros((self.d,), float), self.covariance, size=size
        ))
        indices = random_state.choice(self.n, size=size, p=self.weights)
        means = self.dataset[:, indices]

        return means + norm

    @staticmethod
    def _get_covariance_factor_function_from_bandwidth_type(
            bw_method: tp.Optional[tp.Union[CovarianceFactorFunctionType,
                                            str,
                                            tp.Callable,
                                            numbers.Number]] = None) \
            -> CovarianceFactorFunctionType:
        """
        Infers the bandwidth selection method from
        Args:
            bw_method: either 'scotts' or 'silverman' or a scalar or a function returning a float

        Returns:
        covariance factor function
        """
        if bw_method is None:
            return compute_scotts_factor
        elif isinstance(bw_method, str):
            if bw_method == SCOTTS_FACTOR_STRING:
                return compute_scotts_factor
            elif bw_method == SILVERMAN_FACTOR_STRING:
                return compute_silverman_factor
            else:
                raise ValueError(f'bw_method={bw_method} is not supported')
        elif callable(bw_method):
            return bw_method
        elif np.isscalar(bw_method):
            return lambda kde_info: bw_method
        else:
            raise ValueError(f'bw_method {bw_method} is not supported')

    def _compute_covariance(self):
        """
        Computes the covariance matrix for each Gaussian kernel using
        covariance_factor().
        """
        kde_info = GaussianKDEInformation(dimension=self.d,
                                          n=self.n,
                                          neff=self.neff,
                                          points=self.dataset,
                                          weights=self.weights)

        self.factor = self._covariance_factor(kde_info)

        # Cache covariance and inverse covariance of the data
        if not hasattr(self, '_data_inv_cov'):
            self._data_covariance = atleast_2d(cov(self.dataset, rowvar=1,
                                               bias=False,
                                               aweights=self.weights))
            self._data_inv_cov = linalg.inv(self._data_covariance)

        self.covariance = self._data_covariance * self.factor**2
        self.inv_cov = self._data_inv_cov / self.factor**2
        self._norm_factor = sqrt(linalg.det(2*pi*self.covariance))

    def pdf(self, x: np.ndarray) -> NumericArray:
        """
        Evaluate the estimated pdf on a provided set of points.

        Notes
        -----
        This is an alias for `gaussian_kde.evaluate`.  See the ``evaluate``
        docstring for more details.

        """
        return self.evaluate(x)

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate the log of the estimated pdf on a provided set of points.
        """

        points = atleast_2d(x)

        d, m = points.shape
        if d != self.d:
            if d == 1 and m == self.d:
                # points was passed in as a row vector
                points = reshape(points, (self.d, 1))
                m = 1
            else:
                msg = "points have dimension %s, dataset has dimension %s" % (d,
                    self.d)
                raise ValueError(msg)

        if m >= self.n:
            # there are more points than data, so loop over data
            energy = zeros((self.n, m), dtype=float)
            for i in range(self.n):
                diff = self.dataset[:, i, newaxis] - points
                tdiff = dot(self.inv_cov, diff)
                energy[i] = sum(diff*tdiff, axis=0) / 2.0
            result = logsumexp(-energy.T,
                               b=self.weights / self._norm_factor, axis=1)
        else:
            # loop over points
            result = zeros((m,), dtype=float)
            for i in range(m):
                diff = self.dataset - points[:, i, newaxis]
                tdiff = dot(self.inv_cov, diff)
                energy = sum(diff * tdiff, axis=0) / 2.0
                result[i] = logsumexp(-energy, b=self.weights /
                                      self._norm_factor)

        return result

    @property
    def gpu(self) -> bool:
        return self._gpu

    @property
    def weights(self) -> np.ndarray:
        return self._weights

    @cached_property
    def neff(self) -> float:
        return 1.0/np.sum(self.weights*self.weights)

    @cached_property
    def whitening(self) -> NumericArray:
        gpu = self.gpu
        num_pack = select_num_pack(gpu)
        precision = \
            ensure_consistent_numeric_arrays((self.inv_cov, ), gpu)[0]

        return num_pack.linalg.cholesky(precision)

    @cached_property
    def whitened_points(self) -> NumericArray:
        gpu = self.gpu
        num_pack = select_num_pack(gpu)
        points = \
            ensure_consistent_numeric_arrays((self.dataset.T, ), gpu)[0]

        # print(f'{type(points)=}, {points.dtype=}')
        # print(f'{type(self.whitening)}, {self.whitening.dtype=}')
        return num_pack.dot(points, self.whitening)

    @cached_property
    def normalization_constant(self) -> float:
        gpu = self.gpu
        num_pack = select_num_pack(gpu)

        return (2 * np.pi) ** (- self.d / 2) * num_pack.prod(num_pack.diag(self.whitening))


def evaluate_gaussian_kde_in_batches(kde: gaussian_kde,
                                     points: NumericArray,
                                     maximum_number_of_elements_per_batch: int) \
        -> np.ndarray:
    """
    Evaluates a Gaussian KDE in batches and stores the results in main memory.

    Args:
        kde: a gaussian_kde object
        points:
            numeric array with shape (d, m) containing the points at which to evaluate the kernel
            density estimate
        maximum_number_of_elements_per_batch:
            maximum number of data points times evaluation points to process in a single batch

    Returns:
        a m-dimensional NumPy array of kernel density estimates
    """
    number_of_points = points.shape[1]
    points_per_batch = math.floor(maximum_number_of_elements_per_batch / (kde.n * kde.d))

    n_begin = 0

    output_array = np.zeros((number_of_points, ), dtype=points.dtype)

    while n_begin < number_of_points:
        n_end = min(n_begin + points_per_batch, number_of_points)
        output_array[n_begin:n_end] = np.array(kde.evaluate(points[:, n_begin:n_end]))
        n_begin = n_end

    return output_array



import cocos.device
import cocos.numerics as cn
import numpy as np

from cocos.options import GPUOptions


def test_dot():
    cocos.device.init()

    A_numpy = np.array([[1.0, 2], [3, 4.0]], dtype=np.float32)
    B_numpy = np.array([[5, 6], [7, 8]], dtype=np.float32)
    c_numpy = np.array([[5], [7]], dtype=np.float32)
    d_numpy = c_numpy.transpose()
    e_numpy = np.array([[5, 7]], dtype=np.float32)
    f_numpy = np.array([8, 1], dtype=np.float32)

    # using numpy
    result_1 = np.dot(A_numpy, B_numpy)
    result_2 = np.dot(A_numpy, c_numpy)
    result_3 = np.dot(d_numpy, A_numpy)
    result_4 = np.dot(e_numpy, c_numpy)
    result_5 = np.dot(f_numpy, f_numpy)

    print("numpy result 1")
    print(result_1)

    print("numpy result 2")
    print(result_2)
    print(f"numpy result 2 shape = {result_2.shape}")

    print("numpy result 3")
    print(result_3)

    print("numpy result 4")
    print(result_4)

    print("numpy result 5")
    print(result_5)

    # using Cocos
    A_cocos = cn.array(A_numpy)
    B_cocos = cn.array(B_numpy)
    c_cocos = cn.array(c_numpy)
    d_cocos = cn.array(d_numpy)
    e_cocos = cn.array(e_numpy)
    f_cocos = cn.array(f_numpy)

    print(f"c_cocos.ndim = {c_cocos.ndim}")
    print(f"d_cocos.ndim = {d_cocos.ndim}")
    print(f"e_cocos.ndim = {e_cocos.ndim}")
    print(f"f_cocos.ndim = {f_cocos.ndim}")

    result_1_cocos = cn.dot(A_cocos, B_cocos)
    result_2_cocos = cn.dot(A_cocos, c_cocos)
    result_3_cocos = cn.dot(d_cocos, A_cocos)
    result_4_cocos = cn.dot(e_cocos, c_cocos)
    result_5_cocos = cn.dot(f_cocos, f_cocos)

    print("cocos result 1")
    print(result_1_cocos)

    print("cocos result 2")
    print(result_2_cocos)
    print(f"numpy result 2 shape = {result_2_cocos.shape}")

    print("cocos result 2 reshaped")
    print(np.reshape(result_2_cocos, (2, 1)))

    print("cocos result 3")
    print(result_3_cocos)

    print("cocos result 4")
    print(result_4_cocos)

    print("cocos result 5")
    print(result_5_cocos)

    # conduct tests
    assert np.allclose(result_1, result_1_cocos)
    assert np.allclose(result_3, result_3_cocos)
    assert np.allclose(result_4, result_4_cocos)
    assert np.allclose(result_5, result_5_cocos)
    if GPUOptions.use_gpu:
        assert np.allclose(result_2.flatten(), result_2_cocos)
    else:
        assert np.allclose(result_2.flatten(), result_2_cocos.flatten())



import math
import typing as tp
from enum import Enum


try:
    import pathos
    _IS_PATHOS_INSTALLED = True
except:
    _IS_PATHOS_INSTALLED = False


try:
    import loky
    _IS_LOKY_INSTALLED = True
except:
    _IS_LOKY_INSTALLED = False


def is_loky_installed() -> bool:
    return _IS_LOKY_INSTALLED


def is_pathos_installed() -> bool:
    return _IS_PATHOS_INSTALLED


class MultiprocessingPoolType(Enum):
    LOKY = 1
    PATHOS = 2

    @staticmethod
    def default() -> 'MultiprocessingPoolType':
        if is_loky_installed():
            return MultiprocessingPoolType.LOKY
        elif is_pathos_installed():
            return MultiprocessingPoolType.PATHOS
        else:
            raise ValueError('No suitable multiprocessing package found. '
                             'Please install either loky or pathos.')


def generate_slices_with_batch_size(n: int, batch_size: int) \
        -> tp.Tuple[tp.Tuple[int, int]]:
    """
    Splits range(0, n) into a partition of sub-ranges with length <= batch_size.

    Args:
        n: length of the source range
        batch_size: maximal length of each range in the partition

    Returns:
        tuple of two-element tuples following the pattern (begin_index, end_index)
    """
    begin_index = 0
    result = []

    while begin_index < n:
        end_index = min(begin_index + batch_size, n)
        result.append((begin_index, end_index))
        begin_index += batch_size

    return tuple(result)


def generate_slices_with_number_of_batches(n: int, number_of_batches: int) \
        -> tp.Tuple[tp.Tuple[int, int]]:
    """
    Splits range(0, n) into a partition of number_of_batches sub-ranges of equal length except for the last one.

    Args:
        n: length of the source range
        number_of_batches: the number of sub-ranges in the partition

    Returns:
        tuple of two-element tuples following the pattern (begin_index, end_index)
    """
    batch_size = math.ceil(n/number_of_batches)
    return generate_slices_with_batch_size(n=n, batch_size=batch_size)


ResultType = tp.TypeVar('ResultType')

ParameterTransferFunction = tp.Callable[[tp.Sequence, tp.Dict[str, tp.Any]],
                                        tp.Tuple[tp.Sequence, tp.Dict[str, tp.Any]]]


def _extract_arguments_and_number_of_batches(
        args_list: tp.Optional[tp.Sequence[tp.Sequence]] = None,
        kwargs_list: tp.Optional[tp.Sequence[tp.Dict[str, tp.Any]]] = None,
        number_of_batches: tp.Optional[int] = None):
    if number_of_batches is None:
        if args_list is not None:
            number_of_batches = len(args_list)
        elif kwargs_list is not None:
            number_of_batches = len(kwargs_list)
        else:
            raise ValueError('Number_of_batches must be defined if '
                             'both args_list and kwargs_list are empty')

    if args_list is None:
        args_list = number_of_batches * [list()]
    if kwargs_list is None:
        kwargs_list = number_of_batches * [dict()]

    return args_list, kwargs_list, number_of_batches



import cocos.device
import cocos.numerics as cn
import cocos.scientific.linalg as csl
import numpy as np
import scipy.linalg as scipy_linalg


def compare_cocos_numpy(cocos_array, numpy_array):
    return np.allclose(np.array(cocos_array), numpy_array)


def test_lu():
    cocos.device.init()

    A_numpy = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 20]], dtype=np.float32)

    P, L, U = scipy_linalg.lu(A_numpy)
    # print("P")
    # print(P)
    # print("")
    #
    # print("L")
    # print(L)
    # print("")
    #
    # print("U")
    # print(U)
    # print("")

    A_cocos = cn.array(A_numpy)
    P_cocos, L_cocos, U_cocos = csl.lu(A_cocos)
    # print("P cocos")
    # print(np.array(P_cocos))
    # print("")
    #
    # print("L cocos")
    # print(np.array(L_cocos))
    # print("")
    #
    # print("U cocos")
    # print(np.array(U_cocos))
    # print("")

    assert compare_cocos_numpy(P_cocos, P)
    assert compare_cocos_numpy(L_cocos, L)
    assert compare_cocos_numpy(U_cocos, U)


import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5],
                       [0.4, 0.5, 0.6],
                       [0.7, 0.2, 0.25]],
                      dtype=np.float32),
             # np.array([[0.5, 2.3, 3.1], [4, 5.5, 6], [7 - 9j, 8 + 1j, 2 + 10j]], dtype=np.complex64),
             np.array([[[1.0, 2], [3, 4]],
                       [[5, 6], [7, 8]]],
                      dtype=np.float32)
            ]


@pytest.mark.parametrize("A_numpy", test_data)
def test_ptp(A_numpy):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)

    axes = [None, 0, 1]

    for axis in axes:
        if np.iscomplexobj(A_numpy):
            ptp_cocos = A_cocos.ptp(axis=axis)
            ptp_numpy = A_numpy.ptp(axis=axis)
            print(np.array(ptp_cocos))
            print(np.array(ptp_numpy))
            assert np.allclose(ptp_cocos.real, ptp_numpy.real)
            assert np.allclose(ptp_numpy.imag, ptp_cocos.imag)
        else:
            assert np.allclose(A_cocos.ptp(axis=axis), A_numpy.ptp(axis=axis))



import time
import typing as tp

from cocos.device import (
    ComputeDeviceManager,
    ComputeDevice,
    sync
)

from cocos.multi_processing.utilities import (
    MultiprocessingPoolType,
    ResultType,
    ParameterTransferFunction,
    _extract_arguments_and_number_of_batches
)


def _init_gpu_in_process(device_id: int):
    # print(f'initializing device {device_id}')
    time.sleep(0.1)
    ComputeDeviceManager.set_compute_device(compute_device=device_id)
    # print(f'device {device_id} initialized')


def _get_set_of_compute_devices_from_iterable(
        compute_devices: tp.Iterable[tp.Union[int, ComputeDevice]]) \
        -> tp.FrozenSet[ComputeDevice]:
    result_compute_devices = []

    for i, compute_device in enumerate(compute_devices):
        if isinstance(compute_device, int):
            compute_device \
                = ComputeDeviceManager.get_compute_device(compute_device)
        elif not isinstance(compute_device, ComputeDevice):
            raise TypeError(f"Every element in compute_devices must be of "
                            f"type ComputeDevice or of type int. Entry {i} "
                            f"is of type {type(compute_device)}.")

        result_compute_devices.append(compute_device)

    return frozenset(result_compute_devices)


ComputeDeviceFilter = tp.Callable[[ComputeDevice], bool]


def exclude_intel_devices(compute_device: ComputeDevice) -> bool:
    """
    Some Intel processors feature a GPU integrated on the chip of the CPU. The
    integrated GPU is typically less performant than discrete GPUs in the
    system. This parameter can be used to automatically exclude any device whose
    name contains 'Intel'.
    """

    return 'intel' not in compute_device.name.lower()


class ComputeDevicePool:
    def __init__(self,
                 compute_devices:
                 tp.Optional[tp.Iterable[tp.Union[int, ComputeDevice]]] = None,
                 compute_device_filter:
                 tp.Optional[ComputeDeviceFilter] = exclude_intel_devices,
                 multiprocessing_pool_type: MultiprocessingPoolType = MultiprocessingPoolType.default()) \
            -> None:
        """
        This method constructs a compute device pool from a collection of
        individual devices.

        :param compute_devices: a collection of device ids or compute devices
        :param compute_device_filter: provide a predicate used to filter devices 
                                      to include in the pool
        :param multiprocessing_pool_type: the type of multi-processing pool 
                                          (see class MultiprocessingPoolType)

        """
        if compute_devices is None:
            compute_devices = ComputeDeviceManager.get_compute_devices()

        self._compute_devices \
            = _get_set_of_compute_devices_from_iterable(compute_devices)

        if compute_device_filter is not None:
            compute_devices = \
                    filter(compute_device_filter,
                           [compute_device
                            for compute_device
                            in self._compute_devices])
            self._compute_devices = frozenset(compute_devices)

        if exclude_intel_devices:
            compute_devices = \
                filter(lambda x: 'intel' not in x.name.lower(),
                       [compute_device
                        for compute_device
                        in self._compute_devices])
            self._compute_devices = frozenset(compute_devices)

        # ctx = multiprocessing.get_context("spawn")
        # self._executor = ProcessPoolExecutor(max_workers=self._n_gpus,
        #                                      mp_context=ctx)

        if multiprocessing_pool_type == MultiprocessingPoolType.LOKY:
            from loky import get_reusable_executor, wait

            self._executor = get_reusable_executor(max_workers=self.number_of_devices,
                                                   timeout=None,
                                                   context='loky')

            futures = [self._executor.submit(_init_gpu_in_process,
                                             device_id=compute_device.id)
                       for compute_device
                       in self._compute_devices]

            wait(futures)

            [future.result() for future in futures]
        elif multiprocessing_pool_type == MultiprocessingPoolType.PATHOS:
            from pathos.pools import ProcessPool

            self._executor = ProcessPool(nodes=self.number_of_devices)
            futures = [self._executor.apipe(_init_gpu_in_process, device_id=compute_device.id)
                       for compute_device
                       in self._compute_devices]

            for future in futures:
                while not future.ready():
                    pass
        else:
            raise ValueError(f'Multiprocessing pool type {multiprocessing_pool_type} not supported')

        self._multiprocessing_pool_type = multiprocessing_pool_type

    @property
    def compute_devices(self) -> tp.FrozenSet[ComputeDevice]:
        return self._compute_devices

    @property
    def number_of_devices(self) -> int:
        return len(self.compute_devices)

    @property
    def multiprocessing_pool_type(self) -> MultiprocessingPoolType:
        return self._multiprocessing_pool_type

    def sync(self):
        for compute_device in self._compute_devices:
            compute_device.sync()

    def map_reduce(
            self,
            f: tp.Callable[..., ResultType],
            reduction: tp.Callable[[ResultType, ResultType], ResultType],
            initial_value: ResultType,
            host_to_device_transfer_function:
            tp.Optional[ParameterTransferFunction] = None,
            device_to_host_transfer_function:
            tp.Optional[tp.Callable[[ResultType], ResultType]] = None,
            args_list: tp.Optional[tp.Sequence[tp.Sequence]] = None,
            kwargs_list: tp.Optional[tp.Sequence[tp.Dict[str, tp.Any]]] = None,
            number_of_batches: tp.Optional[int] = None) \
            -> ResultType:

        """
        This method evaluates the function 'f' on elements of 'args_list' and 
        'kwargs_list' in parallel on multiple devices and performs the reduction 
        by calling the function 'reduction' on the result and the result of the 
        reductions so far to eventually produce one final result of type 
        'ResultType'. The reduce step is performed from the left and results are 
        being processed in the same order as they appear in `args_list` and 
        `kwargs_list`. 
    
        Input data to the function f must initially reside in host memory and 
        the user must provide functions 'host_to_device_transfer_function' and 
        'device_to_host_transfer_function' to transfer the data to and results 
        from device memory respectively.
    
        If the arguments for each run of 'f' are identical and they have already 
        been applied to the function that is passed then 'args_list' and 
        'kwargs_list' may both be None but the argument 'number_of_batches' must 
        be specified so the method knows how many times to run the function 'f'.

        Args:
            f: The map function to be evaluated over elements of 'args_list' and 
               'kwargs_list'.
               
            reduction: The reduction to be performed on the results of 'f'. 
                       This is done on the host (not the device).
                       
            initial_value: The initial value of the reduction 
                           (i.e. the neutral element).
                           
            host_to_device_transfer_function: 
                A function that transfers elements of args_list and kwargs_list 
                from host memory to device memory.
                
            device_to_host_transfer_function: 
                A function that transfers results from device to host memory.
                
            args_list: A sequence of sequences of positional arguments.
            kwargs_list: A sequence of dictionaries of keyword arguments.
            number_of_batches: 
                The number of function evaluations is required if 'args_list' 
                and 'kwargs_list' are both empty.

        """

        args_list, kwargs_list, number_of_batches = \
            _extract_arguments_and_number_of_batches(
                args_list=args_list,
                kwargs_list=kwargs_list,
                number_of_batches=number_of_batches)

        def synced_f(index, *args, **kwargs) -> ResultType:
            if host_to_device_transfer_function is not None:
                args, kwargs = host_to_device_transfer_function(*args, **kwargs)
            sync()
            result = f(*args, **kwargs)
            if device_to_host_transfer_function is not None:
                result = device_to_host_transfer_function(result)
            sync()
            return index, result

        results = []
        if self.multiprocessing_pool_type == MultiprocessingPoolType.LOKY:
            from loky import as_completed

            futures = [self._executor.submit(synced_f, i, *args, **kwargs)
                       for i, (args, kwargs)
                       in enumerate(zip(args_list, kwargs_list))]

            for future in as_completed(futures):
                results.append(future.result())
                # result = reduction(result, future.result())
        elif self.multiprocessing_pool_type == MultiprocessingPoolType.PATHOS:
            futures = [self._executor.apipe(synced_f, i, *args, **kwargs)
                       for i, (args, kwargs)
                       in enumerate(zip(args_list, kwargs_list))]

            for future in futures:
                results.append(future.get())
                # result = reduction(result, future.get())
        else:
            raise ValueError(f'Multiprocessing pool type {self.multiprocessing_pool_type} not supported')

        results = sorted(results, key=lambda x: x[0])
        results = [result[1] for result in results]

        result = initial_value
        for new_result in results:
            result = reduction(result, new_result)

        return result

    def map_combine(self,
                    f: tp.Callable[..., ResultType],
                    combination: tp.Callable[[tp.Iterable[ResultType]], ResultType],
                    host_to_device_transfer_function:
                    tp.Optional[ParameterTransferFunction] = None,
                    device_to_host_transfer_function:
                    tp.Optional[tp.Callable[[ResultType], ResultType]] = None,
                    args_list: tp.Optional[tp.Sequence[tp.Sequence]] = None,
                    kwargs_list: tp.Optional[tp.Sequence[tp.Dict[str, tp.Any]]] = None,
                    number_of_batches: tp.Optional[int] = None) -> ResultType:
        """
        This method evaluates the function `f` on elements of `args_list` and 
        `kwargs_list` in parallel on multiple devices and aggregates results 
        in a single step by calling the function `combination` with a list of all 
        results. Results provided to `combination` are in the same order as 
        they appear in `args_list` and `kwargs_list`. 
    
        Input data to the function f must initially reside in host memory and 
        the user must provide functions 'host_to_device_transfer_function' and 
        'device_to_host_transfer_function' to transfer the data to and results 
        from device memory respectively.
    
        If the arguments for each run of 'f' are identical and they have already 
        been applied to the function that is passed then 'args_list' and 
        'kwargs_list' may both be None but the argument 'number_of_batches' must 
        be specified so the method knows how many times to run the function 'f'.

        Args:
            f: The map function to be evaluated over elements of 'args_list' and 
               'kwargs_list'.
               
            combination: 
                A function that aggregates a list of all results in a single step
                
            host_to_device_transfer_function: 
                A function that transfers elements of args_list and kwargs_list 
                from host memory to device memory.
                
            device_to_host_transfer_function:
                 A function that transfers results from device to host memory.
                 
            args_list: A sequence of sequences of positional arguments.
            kwargs_list: A sequence of dictionaries of keyword arguments.
            number_of_batches: 
                The number of function evaluations is required if 'args_list' 
                and 'kwargs_list' are both empty.
        """
        args_list, kwargs_list, number_of_batches = \
            _extract_arguments_and_number_of_batches(
                args_list=args_list,
                kwargs_list=kwargs_list,
                number_of_batches=number_of_batches)

        def synced_f(index, *args, **kwargs) -> ResultType:
            if host_to_device_transfer_function is not None:
                args, kwargs = host_to_device_transfer_function(*args, **kwargs)
            sync()
            result = f(*args, **kwargs)
            if device_to_host_transfer_function is not None:
                result = device_to_host_transfer_function(result)
            sync()
            return index, result

        results = []
        if self.multiprocessing_pool_type == MultiprocessingPoolType.LOKY:
            from loky import as_completed

            futures = [self._executor.submit(synced_f, i, *args, **kwargs)
                       for i, (args, kwargs)
                       in enumerate(zip(args_list, kwargs_list))]

            for future in as_completed(futures):
                results.append(future.result())
        elif self.multiprocessing_pool_type == MultiprocessingPoolType.PATHOS:
            futures = [self._executor.apipe(synced_f, i, *args, **kwargs)
                       for i, (args, kwargs)
                       in enumerate(zip(args_list, kwargs_list))]

            for future in futures:
                results.append(future.get())
        else:
            raise ValueError(f'Multiprocessing pool type {self.multiprocessing_pool_type} not supported')

        results = sorted(results, key=lambda x: x[0])
        results = [result[1] for result in results]

        return combination(results)



import arrayfire as af
import numpy as np
import typing as tp

from ._array import full, ndarray
from ._conversion import convert_numpy_to_af_type
from ._utilities import _pad_shape_tuple_none, _pad_shape_tuple_one


def empty(shape: tp.Tuple[int, ...],
          dtype: np.generic = np.float32) -> ndarray:
    """
    Return a new array of given shape and type, initializing entries to zero.
    """

    return zeros(shape, dtype)


def zeros(shape: tp.Tuple[int, ...],
          dtype: np.generic = np.float32) -> ndarray:
    """
    Return a new array of given shape and type, filled with zeros.
    """

    return full(shape, 0.0, dtype)


def ones(shape: tp.Tuple[int, ...],
         dtype: np.generic = np.float32) -> ndarray:
    """
    Return a new array of given shape and type, filled with ones.
    """

    return full(shape, 1.0, dtype)


def empty_like(a,
               dtype: tp.Optional[np.generic] = None,
               order: str = 'K',
               subok: bool = True) -> ndarray:
    """
    Return a new array with the same shape and type as a given array.
    """

    return zeros_like(a, dtype, order, subok)


def zeros_like(a,
               dtype: tp.Optional[np.generic] = None,
               order: str = 'K',
               subok: bool = True) -> ndarray:
    """
    Return an array of zeros with the same shape and type as a given array.
    """

    if not dtype:
        dtype = a.dtype

    return zeros(a.shape, dtype=dtype)


def ones_like(a,
              dtype: tp.Optional[np.generic] = None,
              order: str = 'K',
              subok: bool = True) -> ndarray:
    """
    Return an array of ones with the same shape and type as a given array.
    """

    if not dtype:
        dtype = a.dtype

    return ones(a.shape, dtype=dtype)


def eye(N: int,
        M: tp.Optional[int] = None,
        k: int = 0,
        dtype: np.generic = np.float32) -> ndarray:
    """
    Return a 2-D array with ones on the diagonal and zeros elsewhere.
    """

    af_type = convert_numpy_to_af_type(dtype)
    if not M:
        M = N

    if k != 0:
        raise ValueError("k not zero is not supported")
    af_array = af.data.identity(N, M, dtype=af_type)

    return ndarray(af_array)


def identity(n: int,
             dtype: np.generic = np.float32) -> ndarray:
    """
    Return the identity array.
    """

    af_type = convert_numpy_to_af_type(dtype)
    af_array = af.data.identity(n, n, dtype=af_type)

    return ndarray(af_array)


def diag(v: ndarray,
         k: int = 0) -> ndarray:
    """
    Extract a diagonal or construct a diagonal array.
    """

    if v.ndim == 1:
        af_array = af.data.diag(v._af_array, num=k, extract=False)
    elif v.ndim == 2:
        af_array = af.data.diag(v._af_array, num=k, extract=True)
    else:
        raise ValueError("v must be a 1d or a 2d ndarray")

    return ndarray(af_array)


def diagflat(v: ndarray,
             k: int = 0):
    """
    Create a two-dimensional array with the flattened input as a diagonal.
    """

    if v.ndim != 1:
        raise ValueError(f"input must be a flat array but has {v.ndim} axes")

    return diag(v, k)


def concatenate(arrays: tp.Tuple[ndarray, ...],
                axis: int = 0) -> ndarray:
    if len(arrays) < 2 or len(arrays) > 4:
        raise ValueError("the number of arrays to be concatenated is supposed "
                         "to be between two and four")

    first = arrays[0]._af_array
    second = arrays[1]._af_array
    third = None
    fourth = None

    if len(arrays) > 2:
        third = arrays[2]._af_array
        if len(arrays) > 3:
            fourth = arrays[3]._af_array

    af_array = af.join(axis, first, second, third, fourth)

    return ndarray(af_array)


def vstack(tup: tp.Tuple[ndarray, ...]) -> ndarray:
    return concatenate(tup, 0)


def hstack(tup: tp.Tuple[ndarray, ...]) -> ndarray:
    return concatenate(tup, 1)


def dstack(tup: tp.Tuple[ndarray, ...]) -> ndarray:
    return concatenate(tup, 2)


def roll(a: ndarray,
         shift: int,
         axis: tp.Optional[int] = None) -> ndarray:
    if axis is None:
        (d0, d1, d2, d3) = _pad_shape_tuple_one(a.shape)
        flat_af_array = af.data.flat(a._af_array)
        af_array_flat_shift = af.data.shift(flat_af_array, shift)
        af_array = af.data.moddims(af_array_flat_shift, d0, d1, d2, d3)
    else:
        shift_dims = [0, 0, 0, 0]
        shift_dims[axis] = shift
        d0, d1, d2, d3 = tuple(shift_dims)

        af_array = af.shift(a._af_array, d0, d1, d2, d3)

    return ndarray(af_array)


def flip(m: ndarray,
         axis: int) -> ndarray:
    return ndarray(af.data.flip(m._af_array, axis))


def fliplr(m: ndarray) -> ndarray:
    return ndarray(af.data.flip(m._af_array, 1))


def flipud(m: ndarray) -> ndarray:
    return ndarray(af.data.flip(m._af_array, 0))


def tril(m: ndarray,
         k: int = 0) -> ndarray:
    """
    Lower triangle of an array.
    """
    af_array = af.data.lower(m._af_array, is_unit_diag=False)

    return ndarray(af_array)


def triu(m: ndarray,
         k: int = 0) -> ndarray:
    """
    Upper triangle of an array.
    """
    af_array = af.data.upper(m._af_array, is_unit_diag=False)

    return ndarray(af_array)



from abc import ABC, abstractmethod
from types import ModuleType
import typing as tp


class NumericalPackageBundle(ABC):
    @classmethod
    @abstractmethod
    def is_installed(cls) -> bool:
        pass

    @classmethod
    @abstractmethod
    def label(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def module(cls) -> ModuleType:
        pass

    @classmethod
    @abstractmethod
    def random_module(cls) -> ModuleType:
        pass

    @classmethod
    def synchronize(cls):
        pass


# class AFNumpyBundle(NumericalPackageBundle):
#     @classmethod
#     def is_installed(cls) -> bool:
#         try:
#             import afnumpy
#             return True
#         except:
#             return False
#
#     @classmethod
#     def label(cls) -> str:
#         return 'afnumpy'
#
#     @classmethod
#     def module(cls) -> ModuleType:
#         import afnumpy
#         return afnumpy
#
#     @classmethod
#     def random_module(cls) -> ModuleType:
#         import afnumpy.random
#         return afnumpy.random
#
#     @classmethod
#     def synchronize(cls):
#         from arrayfire import sync
#         sync()


class NumpyBundle(NumericalPackageBundle):
    @classmethod
    def is_installed(cls) -> bool:
        try:
            import numpy
            return True
        except:
            return False

    @classmethod
    def label(cls) -> str:
        return 'NumPy'

    @classmethod
    def module(cls) -> ModuleType:
        import numpy
        return numpy

    @classmethod
    def random_module(cls) -> ModuleType:
        import numpy.random
        return numpy.random


class CocosBundle(NumericalPackageBundle):
    @classmethod
    def is_installed(cls) -> bool:
        try:
            import cocos
            return True
        except:
            return False

    @classmethod
    def label(cls) -> str:
        return 'Cocos'

    @classmethod
    def module(cls) -> ModuleType:
        import cocos.numerics
        return cocos.numerics

    @classmethod
    def random_module(cls) -> ModuleType:
        import cocos.numerics.random
        return cocos.numerics.random

    @classmethod
    def synchronize(cls):
        from cocos.device import sync
        sync()


class CuPyBundle(NumericalPackageBundle):
    @classmethod
    def is_installed(cls) -> bool:
        try:
            import cupy
            return True
        except:
            return False

    @classmethod
    def label(cls) -> str:
        return 'CuPy'

    @classmethod
    def module(cls) -> ModuleType:
        import cupy
        return cupy

    @classmethod
    def random_module(cls) -> ModuleType:
        import cupy.random
        return cupy.random

    @classmethod
    def synchronize(cls):
        import cupy
        cupy.cuda.Stream.null.synchronize()


def get_available_numerical_packages(
        list_installed_bundles: tp.Optional[bool] = False) \
        -> tp.Tuple[tp.Type[NumericalPackageBundle], ...]:
    numerical_bundles_to_try = (NumpyBundle,
                                CocosBundle,
                                CuPyBundle,
                                # AFNumpyBundle
                                )

    available_numerical_bundles \
        = [numerical_bundle
           for numerical_bundle in numerical_bundles_to_try
           if numerical_bundle.is_installed()]

    if list_installed_bundles:
        print(f'Required packages found for the following benchmarks:')
        for numerical_bundle in available_numerical_bundles:
            print(numerical_bundle.label())

        print()

    return tuple(available_numerical_bundles)



import numpy as np
from scipy.stats import kstest
import typing as tp

import cocos.numerics as cn


def perform_ks_test(x: tp.Union[np.ndarray, cn.ndarray],
                    alpha: float,
                    distribution,
                    args: tp.Tuple = None,
                    verbose: bool = False) -> bool:
    D, p_value = kstest(x, distribution, args=args)
    if verbose:
        print(f"D={D}, p-value={p_value}")

    alpha = 0.01
    reject = p_value < alpha
    if verbose:
        print(f"D={D}, p-value={p_value}")
        print(f"reject = {reject}")

    return reject



from numbers import Number
import numpy as np
import typing as tp

import cocos.numerics as cn

NumericArray = tp.TypeVar('NumericArray',
                          np.ndarray,
                          cn.ndarray)

IntOrNumericArray = tp.TypeVar('IntOrNumericArray',
                               int,
                               np.ndarray,
                               cn.ndarray)

FloatOrNumericArray \
    = tp.TypeVar('FloatOrNumericArray',
                 float,
                 np.ndarray,
                 cn.ndarray)

ComplexOrNumericArray \
    = tp.TypeVar('ComplexOrNumericArray',
                 complex,
                 np.ndarray,
                 cn.ndarray)

ComplexOrFloatOrNumericArray \
    = tp.TypeVar('ComplexOrFloatOrNumericArray',
                 complex,
                 float,
                 np.ndarray,
                 cn.ndarray)

NumericArrayOrScalar = tp.TypeVar('NumericArrayOrScalar',
                                  np.ndarray,
                                  cn.ndarray, Number)



import numpy as np
from numpy.core.multiarray import array

from cocos.numerics._arith import \
    absolute, \
    arccos, \
    arccosh, \
    angle, \
    arcsin, \
    arcsinh, \
    arctan, \
    arctan2, \
    arctanh, \
    ceil, \
    cbrt, \
    cos, \
    cosh, \
    exp, \
    expm1, \
    erf, \
    erfc, \
    factorial, \
    gamma, \
    hypot, \
    log, \
    log10, \
    gammaln, \
    mod, \
    power, \
    root, \
    sign, \
    sin, \
    sinh, \
    sqrt, \
    tan, \
    tanh

from cocos.numerics._array import \
    conj, \
    floor, \
    imag, \
    real

# from cocos.numerics import \
#     arccos, \
#     arccosh, \
#     angle, \
#     arcsin, \
#     arcsinh, \
#     arctan, \
#     arctan2, \
#     arctanh, \
#     ceil, \
#     cbrt, \
#     conj, \
#     cos, \
#     cosh, \
#     exp, \
#     expm1, \
#     erf, \
#     erfc, \
#     factorial, \
#     floor, \
#     gamma, \
#     hypot, \
#     imag, \
#     log, \
#     log10, \
#     gammaln, \
#     mod, \
#     power, \
#     real, \
#     root, \
#     sign, \
#     sin, \
#     sinh, \
#     sqrt, \
#     tan, \
#     tanh

inf = np.inf
COCOS_TRANSLATIONS = {
    "abs": absolute,
    "acos": arccos,
    "acosh": arccosh,
    "arg": angle,
    "asin": arcsin,
    "asinh": arcsinh,
    "atan": arctan,
    "atan2": arctan2,
    "atanh": arctanh,
    "ceiling": ceil,
    "cbrt": cbrt,
    "conjugate": conj,
    "cos": cos,
    "cosh": cosh,
    "exp": exp,
    "E": exp,
    "expm1": expm1,
    "erf": erf,
    "erfc": erfc,
    "fac": factorial,
    "factorial": factorial,
    "floor": floor,
    "gamma": gamma,
    "hypot": hypot,
    "I": 1j,
    "im": imag,
    "ln": log,
    "log": log,
    "log10": log10,
    "loggamma": gammaln,
    "Mod": mod,
    "oo": inf,
    "power": power,
    "re": real,
    "root": root,
    "sign": sign,
    "sin": sin,
    "sinh": sinh,
    "sqrt": sqrt,
    "tan": tan,
    "tanh": tanh,
    "SparseMatrix": array,
    "ImmutableSparseMatrix": array,
    "Matrix": array,
    "MutableDenseMatrix": array,
    "ImmutableMatrix": array,
    "ImmutableDenseMatrix": array,
}



import cocos.numerics as cn
from cocos.numerics.numerical_package_selector import \
    select_num_pack_by_dtype_from_iterable

from cocos.numerics.data_types import NumericArray

from cocos.symbolic import \
    LambdifiedVectorExpression, \
    find_length_of_state_vectors

import numpy as np
import sympy as sym


def jacobian_direct(x1, x2, x3) -> NumericArray:
    state_vectors = (x1, x2, x3)
    R = find_length_of_state_vectors(state_vectors)
    num_pack = select_num_pack_by_dtype_from_iterable(state_vectors)
    result = num_pack.zeros((R, 3, 3))

    result[:, 0, 0] = 1.0
    result[:, 0, 1] = 1.0
    result[:, 0, 2] = 0.0

    result[:, 1, 0] = 2.0 * (x1 + x3)
    result[:, 1, 1] = 0.0
    result[:, 1, 2] = 2.0 * (x1 + x3)

    result[:, 2, 0] = num_pack.exp(x1 + x2)
    result[:, 2, 1] = num_pack.exp(x1 + x2)
    result[:, 2, 2] = 0.0

    return result


def test_lambdified_array_expression():
    # define symbolic arguments to the function
    x1, x2, x3, t = sym.symbols('x1, x2, x3, t')
    argument_symbols = (x1, x2, x3)

    # define a function f: R^3 -> R^3
    f = sym.Matrix([[x1 + x2], [(x1 + x3) ** 2], [sym.exp(x1 + x2)]])

    # Compute the Jacobian symbolically
    jacobian_f = f.jacobian([x1, x2, x3])

    # Convert the symbolic array expression to an object that can evaluated
    # numerically on the cpu or gpu.
    jacobian_f_lambdified \
        = LambdifiedVectorExpression(argument_symbols=argument_symbols,
                                     # time_symbol=t,
                                     symbolic_vector_expression=jacobian_f)

    # Compare the results on cpu and gpu with direct evaluation of the jaboian
    # at n different vectors (x1, x2, x3)
    n = 10000
    print(f'evaluating Jacobian at {n} vectors\n')

    X_gpu = cn.random.rand(n, 3)
    X_cpu = np.array(X_gpu)

    # evaluate on GPU
    jacobian_f_numeric_gpu_direct = \
        jacobian_direct(x1=X_gpu[:, 0],
                        x2=X_gpu[:, 1],
                        x3=X_gpu[:, 2])

    jacobian_f_numeric_gpu_using_matrix = \
        jacobian_f_lambdified.evaluate(X_gpu, t=0)

    assert np.allclose(jacobian_f_numeric_gpu_direct,
                       jacobian_f_numeric_gpu_using_matrix)

    jacobian_f_numeric_gpu_using_kwargs = \
        (jacobian_f_lambdified
         .evaluate_with_kwargs(x1=X_gpu[:, 0],
                               x2=X_gpu[:, 1],
                               x3=X_gpu[:, 2],
                               # t=0
                               )
         )
    assert np.allclose(jacobian_f_numeric_gpu_direct,
                       jacobian_f_numeric_gpu_using_kwargs)

    jacobian_f_numeric_gpu_using_dictionary = \
        (jacobian_f_lambdified
         .evaluate_with_dictionary(
            symbolic_to_numeric_parameter_map={x1: X_gpu[:, 0],
                                               x2: X_gpu[:, 1],
                                               x3: X_gpu[:, 2]},
            # t=0
        )
        )
    assert np.allclose(jacobian_f_numeric_gpu_direct,
                       jacobian_f_numeric_gpu_using_dictionary)

    # evaluate on CPU
    jacobian_f_numeric_cpu_direct = \
        jacobian_direct(x1=X_cpu[:, 0],
                        x2=X_cpu[:, 1],
                        x3=X_cpu[:, 2])

    jacobian_f_numeric_cpu_using_matrix = \
        jacobian_f_lambdified.evaluate(X_cpu, t=0)

    assert np.allclose(jacobian_f_numeric_cpu_direct,
                       jacobian_f_numeric_cpu_using_matrix)

    jacobian_f_numeric_cpu_using_kwargs = \
        (jacobian_f_lambdified
         .evaluate_with_kwargs(x1=X_cpu[:, 0],
                               x2=X_cpu[:, 1],
                               x3=X_cpu[:, 2],
                               # t=0
                               )
         )
    assert np.allclose(jacobian_f_numeric_cpu_direct,
                       jacobian_f_numeric_cpu_using_kwargs)

    jacobian_f_numeric_cpu_using_dictionary = \
        (jacobian_f_lambdified
         .evaluate_with_dictionary(
            symbolic_to_numeric_parameter_map={x1: X_cpu[:, 0],
                                               x2: X_cpu[:, 1],
                                               x3: X_cpu[:, 2]},
            # t=0
        )
        )
    assert np.allclose(jacobian_f_numeric_cpu_direct,
                       jacobian_f_numeric_cpu_using_dictionary)

    # Verify that the results on CPU and GPU match
    assert np.allclose(jacobian_f_numeric_cpu_using_matrix,
                       jacobian_f_numeric_cpu_using_matrix)
